import numpy as np

def get_optimizer(name: str):
    """
    Returns the optimization function corresponding to the provided name.
    """
    if name == "sgd":
        return sgd_update


def sgd_update(parameter: np.ndarray, gradient: np.ndarray, learning_rate: float) -> np.ndarray:
    """
    Updates parameters using Stochastic Gradient Descent (SGD).

    Args:
        parameter (np.ndarray): Current values of the parameters that need to be updated.
        gradient (np.ndarray): Gradients computed during the backpropagation step.
        learning_rate (float): The step size to use during the SGD update.

    Returns:
        np.ndarray: Updated parameters after applying the SGD step.
    """    
    return parameter - learning_rate * gradient



import numpy as np
from typing import Callable, Any

def get_loss_function(loss: str) -> Callable[..., Any]:
    """
    Returns the loss function corresponding to the provided name.
    """
    if loss == "mse":
        return mse


def get_loss_derivative(loss: str) -> Callable[..., Any]:
    """
    Returns the derivative of the loss function corresponding to the provided name.
    """
    if loss == "mse":
        return mse_derivative
    
    
def mse(yhat: float, y: float) -> float:
    """
    Mean Squared Error Function.

    Args:
        yhat (float): Prediction.
        y (float): True label.
    """    
    return np.mean(0.5 * (yhat - y)**2)


def mse_derivative(yhat: float, y: float) -> float:
    """
    Derivative of the Mean Squared Error Function.

    Args:
        yhat (float): Prediction.
        y (float): True Label.
    """    
    return yhat - y


import numpy as np
from typing import Callable, Any

def get_activation_function(name: str) -> Callable[..., Any]:
    """
    Returns the activation function corresponding to the provided name.
    """
    if name == "sigmoid":
        return sigmoid
    
    
def get_activation_derivative(name: str) -> Callable[..., Any]:
    """
    Returns the derivative of the activation function corresponding to the provided name.
    """
    if name == "sigmoid":
        return sigmoid_derivative


def sigmoid(x: float) -> float:
    """
    Sigmoid function.
    """
    return 1.0 / (1.0 + np.exp(-x))


def sigmoid_derivative(x: float) -> float:
    """
    Derivative of the sigmoid function.
    """
    return sigmoid(x) * (1 - sigmoid(x))




import numpy as np
from skynet_ml.nn.initializers import initialize_weights, initialize_bias
from skynet_ml.nn.activations import get_activation_function, get_activation_derivative
from skynet_ml.nn.optimizers import get_optimizer

class Dense:
    """
    Represents a fully connected (dense) neural network layer. Each neuron in this layer is connected to every neuron in the previous layer.
    
    Attributes:
        weights (np.array): Weights matrix of the layer.
        bias (np.array): Bias vector for the layer. Set to zeros if `has_bias=False`.
        activation (function): Activation function for neurons.
        activation_derivative (function): Derivative of the activation function.
        has_bias (bool): Indicator for including bias.
    """    


    def __init__(self, input_size: int, output_size: int, activation: str = "sigmoid", has_bias: bool = True, initialization: str = "random") -> None:
        """
        Initializes the Dense layer.

        Args:
            input_size (int): Number of input features or neurons from the previous layer.
            output_size (int): Number of neurons in this layer.
            activation (str, optional): Activation function's name. Defaults to "sigmoid".
            has_bias (bool, optional): Indicator for bias units. Defaults to True.
            initialization (str, optional): Method to initialize weights and biases. Defaults to "random".
        """
        self.weights = initialize_weights(initialization, input_size, output_size)
        self.bias = initialize_bias(has_bias, initialization, output_size)
        self.activation = get_activation_function(activation)
        self.activation_derivative = get_activation_derivative(activation)
        self.has_bias = has_bias


    def forward(self, input_vector: np.array) -> np.array:
        """
        Computes the forward pass through the layer.

        Args:
            input_vector (np.array): Layer inputs.

        Returns:
            np.array: Activations after linear transformation and activation function.
        """
        # Store input for backpropagation
        self.input_vector = input_vector 
        # Linear transformation (Wx + b)
        self.z = np.dot(self.weights, self.input_vector) + self.bias
        # Activation(Wx + b)
        self.a = self.activation(self.z)
        
        return self.a


    def backward(self, d_output: np.array) -> np.array:
        """
        Computes the gradients using backpropagation.

        Args:
            d_output (np.array): Gradient of the loss with respect to the outputs of this layer. It's the first term of the delta for this layer.

        Returns:
            np.array: Gradient w.r.t. layer inputs.
        """
        # Calculate delta: combining upstream gradient and gradient of the activation (d_L/d_a * d_a/d_z)
        self.delta = np.multiply(d_output, self.activation_derivative(self.z)) 
        # Gradient w.r.t. weights
        self.d_weights = np.dot(self.delta, self.input_vector.T)
        # Gradient w.r.t. bias (if present)
        self.d_bias = np.sum(self.delta, axis=1, keepdims=True) if self.has_bias else None
        # Gradient w.r.t. layer inputs for backpropagation to the previous layer (d_L/d_a for next layer)
        d_input = np.dot(self.weights.T, self.delta)
        
        return d_input


    def update_parameters(self, optimizer: str = "sgd", learning_rate: float = 0.01) -> None:
        """
        Updates parameters (weights and biases) using the provided optimizer.

        Args:
            optimizer (str, optional): Optimization method. Defaults to "sgd".
            learning_rate (float, optional): Step size for optimizer. Defaults to 0.01.
        """
        # Fetch optimizer function
        optimizer_function = get_optimizer(optimizer)
        
        # Update weights
        self.weights = optimizer_function(self.weights, self.d_weights, learning_rate)
        # Update bias if present
        if self.has_bias:
            self.bias = optimizer_function(self.bias, self.d_bias, learning_rate)



import numpy as np

def initialize_weights(method: str, input_size: int, output_size: int) -> np.array:
    """
    Initializes the weights based on the initialization method.

    Args:
        method (str): Initialization method.
        input_size (int): Number of columns in the weight matrix.
        output_size (int): Number of rows in the weight matrix.

    Returns:
        np.array: weight matrix.
    """    
    if method == 'random':
        return np.random.randn(output_size, input_size)


def initialize_bias(has_bias: bool, method: str, output_size: int) -> np.array:
    """
    Initializes the bias based based on the initialization method. 

    Args:
        has_bias (bool): If bias should have value or not.
        method (str): Initialization method.
        output_size (int): Number of values in bias vector.

    Returns:
        np.array: Bias vector.
    """    
    if has_bias:
        if method == "random":
            return np.random.rand(output_size, 1)
            
    return np.zeros((output_size, 1))



import numpy as np
from typing import List, Tuple
from skynet_ml.nn.losses import get_loss_function, get_loss_derivative

class Sequential:
    """
    Represents a sequential neural network. 
    
    Attributes:
        epochs (int): Number of epochs the model will be trained with.
        learning_rate (float): Learning rate for the optimizer.
        optimizer (str): Optimization algorithm to use.
        batch_size (int): Size of each mini_batch.
        loss_function (function): Loss function to be used.
        loss_derivative (function): Derivative of the loss function.
        loss_epochs (list): List with the loss for each epoch.
    """    
    
    def __init__(self) -> None:
        """
        Initializes the Sequential model.
        
        Attributes: 
        - layers: List of layers added to the neural network.
        """        
        self.layers: list = []
        
        
    def add(self, layer) -> None:
        """
        Add a new layer to the neural network.

        Args:
            layer: The layer to be added.
        """        
        self.layers.append(layer)
        
        
    def compile(self, epochs: int, learning_rate: float, optimizer: str, batch_size: int, loss: str) -> None:
        """Configure the learning process before training starts.

        Args:
            epochs (int): Number of epochs the model will be tranied with.
            learning_rate (float): Learning rate for the optimizer.
            optimizer (str): Optimization algorithm to use.
            batch_size (int): Size of each mini batch.
            loss (str): Loss function to be used.
        """        
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.optimizer = optimizer
        self.batch_size = batch_size
        self.loss_function = get_loss_function(loss)
        self.loss_derivative = get_loss_derivative(loss)
        self.loss_epochs = []
        
        
    def forward(self, X: np.array) -> np.array:
        """
        Performs the forward pass through the network.

        Args:
            X (np.array): Input data.

        Returns:
        - X (np.array): yhat vector. 
        """        
        for layer in self.layers:
            X = layer.forward(X)
        return X
    
    
    def backward(self, d_output: np.array) -> None:
        """Perform the backward pass (backpropagation) through all the layers of the network

        Args:
            d_output (np.array): First term of the delta fo each layer.
        """        
        for layer in reversed(self.layers):
            d_output = layer.backward(d_output) # gradient w.r.t input
    
            
    def optimize(self) -> None:
        """
        Updates the weights and biases for each layer.
        """        
        for layer in self.layers:
            layer.update_parameters(self.optimizer, self.learning_rate)
    
            
    def compute_loss(self, yhat: np.array, y: np.array) -> np.array:
        """
        Computes the loss of the networks.

        Args:
            yhat (np.array): Predictions (output of the network).
            y (np.array): True labels.

        Returns:
            np.array: The vector with the loss values.
        """        
        return self.loss_function(yhat, y)
    
    
    def compute_loss_derivative(self, yhat: np.array, y: np.array) -> np.array:
        """
        Computes the derivative of the loss w.r.t the predictions.

        Args:
            yhat (np.array): Predictions
            y (np.array): True Labels.

        Returns:
            np.array: The vector with the derivative of the loss values.
        """        
        return self.loss_derivative(yhat, y)
    
    
    def create_mini_batches(self, X: np.array, y: np.array, batch_size: int) -> List[Tuple[np.array, np.array]]:
        """
        Create mini-batches from the provided data.

        Args:
        - X: Input data.
        - y: True labels.
        - batch_size: Desired size of each mini-batch.

        Returns:
        - mini_batches: List of tuples, where each tuple contains a mini-batch of data and corresponding labels.
        """
        # Create an array of indices from 0 to the number of samples.
        indices = np.arange(X.shape[1])
        np.random.shuffle(indices)  # Shuffle the indices.
        X = X[:, indices]  # Shuffle X using the shuffled indices.
        y = y[:, indices]  # Shuffle y using the shuffled indices.

        mini_batches = []

        total_batches = X.shape[1] // batch_size
        for i in range(total_batches):
            X_mini = X[:, i * batch_size: (i + 1) * batch_size]
            y_mini = y[:, i * batch_size: (i + 1) * batch_size]
            mini_batches.append((X_mini, y_mini))

        # Handle the end case (last mini-batch < mini_batch_size)
        if X.shape[1] % batch_size != 0:
            X_mini = X[:, total_batches * batch_size:]
            y_mini = y[:, total_batches * batch_size:]
            mini_batches.append((X_mini, y_mini))

        return mini_batches
    
    
    def fit(self, X: np.array, y: np.array) -> None:
        """
        Train the neural network using the provided data.

        Args:
            X (np.array): Input data for training.
            y (np.array): True labels for the input data.
        """        
        for i in range(self.epochs):
            # create the mini batches for training
            mini_batches = self.create_mini_batches(X, y, self.batch_size)
            # list to store the loss for each mini batch
            loss_batches = []
            
            for batch in mini_batches:
                # unpack mini batch data and labels
                X_mini, y_mini = batch
                # forward pass: compute predictions
                yhat_mini = self.forward(X_mini)
                # compute loss of the mini batch
                loss_mini = self.compute_loss(yhat_mini, y_mini)
                # compute derivative of loss of the mini batch
                loss_derivative_mini = self.compute_loss_derivative(yhat_mini, y_mini)
                # store the loss
                loss_batches.append(loss_mini)
                # backward pass: compute the gradient of the loss w.r.t each weight and bias
                self.backward(loss_derivative_mini)
                # optimize the network's weights and bias
                self.optimize()
            
            # average loss for the epoch
            epoch_loss = np.mean(loss_batches)
            print(f"Epoch {i} Loss: {epoch_loss}")
            self.loss_epochs.append(epoch_loss)
            
            
    def predict(self, X: np.array) -> np.array:
        """
        Use the network to predict a given a given input matrix X.

        Args:
            X (np.array): Input data for prediction

        Returns:
            np.array: The predicted values.
        """        
        return self.forward(X)